;-------------------------------------------------------------------------------
;Streaming Lempel-Ziv decompressor.
;
;The decoder renders one page of output per call, and potentially even writes
;parts of the subsequent page. It is entirely up to the caller to advance the
;lz_dst msb (or not).
;
;The point of this scheme is that several compressed blocks or other data can be
;packed end-to-end instead of requiring things to be aligned on sector
;boundaries
;
;Finally there's a crude but effective way to use sliding-windows. Recall that
;the caller is responsible for advancing the destination pointer, and thus also
;needs wrap it when necessary. furthermore the compressor has to be told of a
;limited window-size so it doesn't try to use matches beyond it or, and this is
;an important point, matches spanning the window boundary.
;
;Then the one thing that's left for the decruncher to do is to wrap backwards
;references when subtracting the offset from the current output position.
;The easiest way to achieve this is to make the window a nice even power-of-two
;and place it on a an odd address, then fix up the high-byte with a single ORA.
;In other words ORA #$10 would work for a 4k window placed at
;$1000/$3000/$5000/$7000/$9000/$a000/$f000.
;-------------------------------------------------------------------------------
;lz_dst		= $fb			;Decompression destination pointer.
;					;Initialize this to whatever address
;					;you want to decompress to
;
;lz_bits	= $fd			;Internal shift register. Must be
;					;preserved between lz_decrunch class
;					;and initialized to $80 on startup
;
;lz_scratch	= $fe			;Temporary zeropage storage.
;					;Need *not* be preserved between
;					;lz_decrunch calls
;
;lz_sector	= $0400			;The one-page buffer from which the
;					;compressed data is actually read,
;					;and which gets refilled by
;					;lz_fetch_sector


;-------------------------------------------------------------------------------
;This is the user's hook to replenish the sector buffer with some new bytes.
;
;A and Y are expected to be preserved while carry must be set on exit.
;X should point to the first byte of the new data, e.g. zero for a full 256-byte
;page of data or two to skip past the sector and track links.
;
;When fetching from a larger in-memory array rather than a single sector buffer
;the lz_sector_ptr1..3 pointers will need to be patched up
;-------------------------------------------------------------------------------
;lz_fetch_sector
;		pha			;Preserve A/Y
;		tya
;		pha
;
;		jsr loader		;Run your loader to fill up lz_sector
;
;		ldx #$00		;Rewind the sourcepointer
;		pla
;		tay
;		pla
;		sec			;Always raise carry on exit
;		rts


;-------------------------------------------------------------------------------
;Typical usage
;-------------------------------------------------------------------------------
;		;Prefetch the initial sector. This also takes care of reloading
;		;the X offset
;		jsr lz_fetch_sector
;
;		lda #$80		;Mark the shift register as empty
;		sta lz_bits
;
;		lda #<window		;Start writing to the beginning of the
;		sta lz_dst+0		;destination window
;		lda #>window
;_lz_unpack	sta lz_dst+1
;
;		jsr lz_decrunch		;Decode the next page of input
;
;		;Either process the data here or continue straight away unless
;		;getting dangerously close to an interrupt-based consumer
;
;		inc lz_dst+1		;Increment and wrap the window pointer
;		lda lz_dst+1
;		ora #$nn
;		bcs _lz_unpack		;Continue until reaching the end of the
;					;file


;-------------------------------------------------------------------------------
;This is the main lz_decrunch function which fills up the rest of the current
;lz_dst page in the decompression window along with possible parts of the next.
;
;On entry and exit the X register points to the next available byte in the
;sector buffer, in ascending order from $00 to $ff.
;This implies that the initial sector must have already been fetched, and that a
;file ending with X wrapped to $00 will have needlessly fetched an extra sector
;(which may be taken advantage of when decoding a contiguous set of files.)
;
;Carry will be cleared on exit upon reaching the end of the file
;-------------------------------------------------------------------------------

		;******** Fetch some more bits to work with ********

lz_sector_ptr1	= *+1
_lz_refill_bits	ldy lz_sector,x
		sty lz_bits
		inx
		bne *+5
		jsr lz_fetch_sector
;		sec
		rol lz_bits
_lz_upper_rts	rts


		;******** Start the next match/literal run ********

_lz_lcopy_fetch	jsr lz_fetch_sector	;Grab a new sector for the literal loop
		bcs _lz_lcopy_cont	;(BRA)

		;Literal or match to follow?
_lz_type_refill	jsr _lz_refill_bits
		bne *+4			;Alternatively save a byte by skipping
;		.byte $0C		;the next instruction through an illegal
					;NOP
lz_decrunch	asl lz_bits
		bcc _lz_do_match
		beq _lz_type_refill


		;******** Process literal run ********

		lda #%00000000		;Decode run length
_lz_lrun_loop	rol
		asl lz_bits
		bcs _lz_lrun_test
_lz_lrun_back	asl lz_bits
		bne _lz_lrun_loop

		jsr _lz_refill_bits
		bne _lz_lrun_loop	;(BRA)

_lz_lrun_test	bne _lz_lrun_gotten
		jsr _lz_refill_bits
		bcc _lz_lrun_back

_lz_lrun_gotten	tay			;Valid lenghts are 1-256, the latter
;		sec			;of which causes trouble for the 8-bit
		sbc #$01		;destination pointer addition.
		sta lz_scratch		;We work around this by pre-decrementing
					;and adding with carry instead

lz_sector_ptr2	= *+1			;Copy the literal data. Note the
_lz_lcopy	lda lz_sector,x		;reverse order of the source
		inx			;characters
		beq _lz_lcopy_fetch
_lz_lcopy_cont	dey
		sta (lz_dst),y
		bne _lz_lcopy

		;Time to advance the destination pointer.
		;Normally the decrement before _lz_lcopy leaves carry set,
		;except in the 256-byte case. However this case also requires
		;a sector refill, and lz_fetch_sector is obligated to raise
		;carry for us
		lda lz_scratch
;		sec
		adc lz_dst+0
		sta lz_dst+0
		bcs _lz_upper_rts

		;One literal run following another only makes sense if the first
		;run is of maximum length and had to be split. Yet any 256-byte
		;run must by definition cross the page, so here we can safely
		;omit the type bit


		;******** Process match ********

_lz_do_match	lda #%00100000		;Determine offset length by a two-bit
_lz_moff_range	asl lz_bits		;prefix combined with the first run
		bne *+5			;length bit (where a one identifies
		jsr _lz_refill_bits	;a two-byte match).
		rol			;The rest of the length bits will
		bcc _lz_moff_range	;then follow *after* the offset data

		tay
		lda _lz_moff_length,y
		beq _lz_moff_far

_lz_moff_loop	asl lz_bits		;Load partial offset byte
		bne *+9
		sty lz_scratch
		jsr _lz_refill_bits
		ldy lz_scratch

		rol
		bcc _lz_moff_loop

		bmi _lz_moff_near

_lz_moff_far	sta lz_scratch		;Save the bits we just read as the
					;high-byte

lz_sector_ptr3	= *+1
		lda lz_sector,x		;For large offsets we can load the
		inx			;low-byte straight from the stream
		bne *+5			;without going throught the shift
		jsr lz_fetch_sector	;register

;		sec
		adc _lz_moff_adjust_lo,y
		bcs _lz_moff_pageok
		dec lz_scratch
		sec
_lz_moff_pageok	adc lz_dst+0
		sta _lz_match+0

		lda lz_scratch
		adc _lz_moff_adjust_hi,y
		sec
		bcs _lz_moff_join	;(BRA)

_lz_moff_near
;		sec			;Special case handling of <8 bit offsets.
	 	adc _lz_moff_adjust_lo,y;We may can safely ignore the MSB from
;		sec			;the base adjustment table as the
		adc lz_dst+0		;maximum base (for a 4/5/6/7 bit
		sta _lz_match+0		;length sequence) is 113
		lda #$ff		;   ____
_lz_moff_join	adc lz_dst+1		;  /   /
					; /   /___
;lz_window	= *+1			;/        | Insert sliding-window
;		ora #$nn		;\    ____| wrapping code here
					; \   \
		sta _lz_match+1		;  \___\


		cpy #$04		;Get any remaning run length bits
		lda #%00000001
		bcs _lz_mrun_gotten

_lz_mrun_loop	asl lz_bits
		bne *+5
		jsr _lz_refill_bits
		rol
		asl lz_bits
		bcc _lz_mrun_loop
		bne _lz_mrun_gotten
		jsr _lz_refill_bits
		bcc _lz_mrun_loop

_lz_mrun_gotten	tay			;A 257-byte (=>$00) run serves as a
		beq _lz_end_of_file	;sentinel

		sta _lz_mcopy_len

		ldy #$ff		;The copy loop. This needs to be run
_lz_mcopy	iny			;forwards since RLE-style matches can
_lz_match	= *+1			;overlap the destination
		lda $ffff,y
		sta (lz_dst),y
_lz_mcopy_len	= *+1
		cpy #$ff
		bne _lz_mcopy

		tya			;Advance destination pointer
;		sec
		adc lz_dst+0
		sta lz_dst+0
		bcs _lz_lower_rts
		jmp lz_decrunch

		;Return temporarily after crossing a page or permanently at the
		;end of the file
_lz_end_of_file	clc
_lz_lower_rts	rts


		;******** Offset coding tables ********

		;This length table is a bit funky. The idea here is to use the
		;value as the initial value of the shift register instead of
		;keeping a separate counter.
		;In other words we iterate until the leading one is shifted out.
		;Then afterwards the bit just below it (our new sign bit) is set
		;if the offset is shorter than 8-bits, and conversely it's
		;cleared if we need to fetch a separate low-byte
		;as well.
		;The fact that the sign bit is cleared as a flag is compensated
		;for in the lz_moff_adjust_hi table

_lz_moff_length
		;Long (>2 byte matches)
		.byte %00011111		;4 bits
		.byte %00000011		;7 bits
		.byte %01011111		;10 bits
		.byte %00001011		;13 bits
		;Short (2 byte matches)
		.byte %01011111		;10 bits
		.byte %00000000		;8 bits
		.byte %00000111		;6 bits
		.byte %00111111		;3 bits
_lz_moff_adjust_lo
		;Long (>2 byte matches)
		.byte %11111110		;1-16
		.byte %11101110		;17-144
		.byte %01101110		;145-1168
		.byte %01101110		;1169-9360
		;Short (2 byte matches)
		.byte %10110110		;329-1352
		.byte %10110110		;73-328
		.byte %11110110		;9-72
		.byte %11111110		;1-8
_lz_moff_adjust_hi = *-2
		;Long (>2 byte matches)
;		.byte %11111111		;1-16 (unreferenced)
;		.byte %11111111		;17-144 (unreferenced)
		.byte %01111111		;145-1168
		.byte %01111011		;1169-9360
		;Short (2 byte matches)
		.byte %01111110		;329-1352
		.byte %11111110		;73-328
;		.byte %11111111		;9-72 (unreferenced)
;		.byte %11111111		;1-8 (unreferenced)
